#include <utility>
#include <list>
#include <cassert>
#include "../../dxlib_vc/Dxlib.h"

// x̗}
#pragma warning(disable:4351) // warning C4351: V: z 'Key::mKeyState' ̗vf͊ŏ܂ => zero-filled

// Macroes
// #define DEBUG
#define SHOULD_NOT_REACH_HERE assert(false && "should not reach here")
#define DISALLOW_COPY_AND_MOVE(className) \
	className(const className&) = delete; \
	className(const className&&) = delete; \
	className& operator=(const className&) = delete; \
	className& operator=(const className&&) = delete;

// Grobal Constants
enum{
	PIECEX = 32,
	PIECEY = 32,
	FIELDX = 8, // ۂ̃{[h
	FIELDY = 14,
	BOARDX = 6, // ڂɌ{[h
	BOARDY = 12, 
};

// Structs
struct Block{
	Block() :x(0), y(0), color(0) {}
	Block(int ax, int ay, int acolor) :x(ax), y(ay), color(acolor) {}
	int x, y, color; // color[1, 4]
};

// Alias
using uint = unsigned int;

//---------------------------------------------------------
// L[{[h֘ANX
//---------------------------------------------------------
class Key final{
private:
	Key()
		:mKeyState{}
	{
	}

public:
	~Key() = default;
	DISALLOW_COPY_AND_MOVE(Key);

	static Key* get()
	{
		static Key inst;
		return &inst;
	}

	// ԍXV
	void Update()
	{
		static char keystateBuf[KEYMAX] = {};
		// keystate
		GetHitKeyStateAll(keystateBuf);
		for (int i = 0; i < KEYMAX; ++i){
			mKeyState[i] = keystateBuf[i] ? ++mKeyState[i] : 0;
		}
	}

	bool Hit(const int keycode) const { return mKeyState[keycode] == 1; }	// uԂǂ
	uint Hold(const int keycode) const { return mKeyState[keycode]; }	// t[Ԃ

private:
	enum{
		KEYMAX = 256,
	};
	uint mKeyState[KEYMAX];
};

//---------------------------------------------------------
// Q[CNX
//---------------------------------------------------------
class Game final{
private:
	Game()
		:mField{}, mNow{}
	{
		this->LoadImg();
		this->InitParams();
	}
public:
	~Game() = default;
	DISALLOW_COPY_AND_MOVE(Game);

	static Game* get()
	{
		static Game inst;
		return &inst;
	}
	void Run();                                 // Q[̎s 

private:
	void LoadImg();                             // 摜ǂݍ
	void InitParams();                          // ϐ
	void NewBlock();                            // VubN~点
	void KeyProcess();                          // L[͏
	void BlockMove(const int dir);              // ubNEɓ
	void Rotate();                              // ]
	void BlockDown();                           // ubN𗎂Ƃ
	void BlockDrop();                           // ubNCɗƂ
	void EraseAll();                            // 4ȏȂubNSď
	std::list<Block> EnumAdjoin();              // ׂ̂̍ẄꗗԂ
	int CountAdjoin(const int x, const int y);	// 
	void CountAdjoinImpl(int fieldBuf[][FIELDX], const int x, const int y, uint* countBuf);	// ċAIɗאړFubN𐔂
	void EraseAdjoin(const int x, const int y);	// אڂ̂
	void FieldDrop();                           // Ŝ𗎉
	void Draw() const;                          // `揈

private:
	// Constants
	enum class GameFlag{
		Normal,
		BlockDelete,
		FieldDrop,
	};

	enum BoardState{
		Board_Wall = -1,
		Board_Empty = 0,
	};

	enum{
		DXLIB_ERROR = -1,         // DXCuŃG[Ƃ̖߂l
		APPEAR_X = 3,             // ubN~ĂXW
		BLOCK_DOWN_FRAME = 30,    // ŃubNĂԊu
	};

	// Local Variables
	int mFrame = 0;                           // o߃t[
	int mField[FIELDY][FIELDX];               // Q[{[h̏
	int mLastStateFrame = 0;                  // Ō̏Ԃ̎ԁiE莞ԂƂɂpj
	bool mIsOver = false;                     // Q[I[o[tO
	Block mNow[2];                            // ݂2̃ubN
	GameFlag mGameFlag = GameFlag::Normal;    // ʏEubNÊǂꂩ̏

	// Images
	using ImageHandle = int;
	ImageHandle mImgPiece[5];
	ImageHandle mImgBg = -1;
};

//---------------------------------
// 摜ǂݍ
void Game::LoadImg()
{
	const char* failedToLoadImgMsg = "摜̓ǂݍ݂Ɏs܂BQ[I܂B";

	if (DXLIB_ERROR == LoadDivGraph("img/puyo.png", 5, 5, 1, PIECEX, PIECEY, mImgPiece))
		throw failedToLoadImgMsg;
	if (DXLIB_ERROR == (mImgBg = LoadGraph("img/background.png")))
		throw failedToLoadImgMsg;
}

//---------------------------------
// ϐ
void Game::InitParams()
{
	this->NewBlock();
	// ǂ
	for (int i = 0; i < FIELDY; ++i){
		mField[i][0] = mField[i][FIELDX - 1] = Board_Wall;
	}
	for (int i = 0; i < FIELDX; ++i){
		mField[FIELDY - 1][i] = Board_Wall;
	}
}

//---------------------------------
// VubN~点
void Game::NewBlock()
{
	mNow[0] = Block(APPEAR_X, 0, GetRand(3) + 1);
	mNow[1] = Block(APPEAR_X, 1, GetRand(3) + 1);
}

//---------------------------------
// L[͏
void Game::KeyProcess()
{
	if (Key::get()->Hit(KEY_INPUT_A))
		this->BlockMove(-1);

	else if (Key::get()->Hit(KEY_INPUT_D))
		this->BlockMove(+1);

	else if (Key::get()->Hit(KEY_INPUT_J))
		this->Rotate();

	else if (Key::get()->Hit(KEY_INPUT_S))
		this->BlockDown();

	else if (Key::get()->Hit(KEY_INPUT_K))
		this->BlockDrop();
}

//---------------------------------
// ubNEɓ
// arg: +1 = right -1 = left
void Game::BlockMove(const int dir)
{
	// ȂΈړAΓȂ
	if (!mField[mNow[0].y][mNow[0].x + dir] &&
		!mField[mNow[1].y][mNow[1].x + dir]){
		mNow[0].x += dir;
		mNow[1].x += dir;
	}
}

//---------------------------------
// ]
void Game::Rotate()
{
	int rot;
	// yɍ => xiEjւ̉]
	if (rot = mNow[1].y - mNow[0].y){
		mNow[1].y -= rot;
		mNow[1].x -= rot;
		// ԂȂ
		if (mField[mNow[1].y][mNow[1].x]){
			// ΂ԂȂ1]
			if (mField[mNow[1].y][mNow[1].x + rot * 2]){
				this->Rotate();
				// yW
				mNow[0].y += rot;
				mNow[1].y += rot;
				return;
			}
			// 󂢂Ă΂炷
			else{
				mNow[0].x += rot;
				mNow[1].x += rot;
			}
		}
	}
	// xɍ => yicjւ̉]
	else if (rot = mNow[1].x - mNow[0].x){
		mNow[1].y += rot;
		mNow[1].x -= rot;
		// ɂԂ1グ
		if (mNow[1].y >= FIELDY - 1){
			--mNow[0].y;
			--mNow[1].y;
		}
	}
}

//---------------------------------
// ubN𗎂Ƃ
void Game::BlockDown()
{
	++mNow[0].y;
	++mNow[1].y;
	// ubNɂԂΎ~߂
	if (mField[mNow[0].y][mNow[0].x] ||
		mField[mNow[1].y][mNow[1].x]){
		// ʂɐ؂藣ėƂ
		this->BlockDrop();
	}
}

//---------------------------------
// ubNCɗƂ
void Game::BlockDrop()
{
	// ɂubNɗƂ Ȃ ɂzW܂ŗĂ܂
	const int lower = mNow[0].y < mNow[1].y;
	const int upper = !lower;
	// Ԃ܂ŉ~E1߂Ċm
	while (mField[mNow[lower].y][mNow[lower].x] == Board_Empty) ++mNow[lower].y;
	mField[--mNow[lower].y][mNow[lower].x] = mNow[lower].color;

	while (mField[mNow[upper].y][mNow[upper].x] == Board_Empty) ++mNow[upper].y;
	mField[--mNow[upper].y][mNow[upper].x] = mNow[upper].color;

	// ΏtO𗧂Ă
	if (this->CountAdjoin(mNow[0].x, mNow[0].y) >= 4 ||
		this->CountAdjoin(mNow[1].x, mNow[1].y) >= 4){
		mGameFlag = GameFlag::BlockDelete;
		mLastStateFrame = GetNowCount();
	}
	else{
		// ̍WɃubNĂ܂Q[I[o[
		if (mField[1][APPEAR_X]){
			mIsOver = true;
			return;
		}
		this->NewBlock();
	}
}

//---------------------------------
// 4ȏȂubNSď
void Game::EraseAll()
{
	auto ignitionPoints = this->EnumAdjoin();
	if (!ignitionPoints.empty()){
		for (const auto& elem : ignitionPoints){
			this->EraseAdjoin(elem.x, elem.y);
		}
		mGameFlag = GameFlag::FieldDrop;
		mLastStateFrame = GetNowCount();
		return;
	}

	// Ԋûߋ؂܂ő҂
	if (mFrame % BLOCK_DOWN_FRAME == 0){
		this->NewBlock();
		mGameFlag = GameFlag::Normal;
	}
}

//---------------------------------
// ׂ̂̒Γ_̍ẄꗗԂ
std::list<Block> Game::EnumAdjoin()
{
	//@pɕ
	int fieldBuf[FIELDY][FIELDX] = {};
	for (int i = 0; i < FIELDY; ++i){
		for (int j = 0; j < FIELDX; ++j){
			fieldBuf[i][j] = mField[i][j];
		}
	}

	// Γ_
	std::list<Block> ignitionPoints;
	for (int i = 1; i < FIELDY - 1; ++i){
		for (int j = 1; j < FIELDX - 1; ++j){
			uint count = 0;
			this->CountAdjoinImpl(fieldBuf, j, i, &count);
			if (count >= 4){
				ignitionPoints.emplace_back(j, i, fieldBuf[i][j]);
			}
		}
	}
	return std::move(ignitionPoints);
}

//---------------------------------
// 
int Game::CountAdjoin(const int x, const int y)
{
	//@pɕ
	int fieldBuf[FIELDY][FIELDX] = {};
	for (int i = 0; i < FIELDY; ++i){
		for (int j = 0; j < FIELDX; ++j){
			fieldBuf[i][j] = mField[i][j];
		}
	}

	// 
	uint count = 0;
	this->CountAdjoinImpl(fieldBuf, x, y, &count);
	return count;
}

//---------------------------------
// ċAIɗאړFubN𐔂
// ۂ̃{[hnΎۂɏ
// arg: tB[h, x, y, אڃJEgϐ̃AhX
void Game::CountAdjoinImpl(int fieldBuf[][FIELDX], const int x, const int y, uint* countBuf)
{
	const int color = fieldBuf[y][x];
	// ǂ󔒂Ȃ炨܂
	if (color <= Board_Empty)
		return;

	fieldBuf[y][x] = Board_Empty;
	if (countBuf != nullptr)
		++*countBuf;

	if (fieldBuf[y - 1][x] == color) this->CountAdjoinImpl(fieldBuf, x, y - 1, countBuf);
	if (fieldBuf[y + 1][x] == color) this->CountAdjoinImpl(fieldBuf, x, y + 1, countBuf);
	if (fieldBuf[y][x - 1] == color) this->CountAdjoinImpl(fieldBuf, x - 1, y, countBuf);
	if (fieldBuf[y][x + 1] == color) this->CountAdjoinImpl(fieldBuf, x + 1, y, countBuf);
}

//---------------------------------
// אڂ̂
void Game::EraseAdjoin(const int x, const int y)
{
	this->CountAdjoinImpl(mField, x, y, nullptr);
}

//---------------------------------
// Ŝ𗎉
void Game::FieldDrop()
{
	for (int i = 1; i < FIELDX; ++i){
		int tmp = FIELDY - 1;
		for (int j = FIELDY - 1; j >= 1; --j){
			if (mField[j][i]){
				mField[tmp--][i] = mField[j][i];
			}
		}
		// cEmpty
		while (tmp >= 0)
			mField[tmp--][i] = Board_Empty;
	}
	mGameFlag = GameFlag::BlockDelete;
	mLastStateFrame = GetNowCount();
}

//---------------------------------
// `揈
void Game::Draw() const
{
	// wi
	DrawGraph(0, 0, mImgBg, true);
	// ݂̃ubN`(͎`悵]r̂悤ɎcĂ܂)
	if (mGameFlag == GameFlag::Normal){
		for (const auto& elem : mNow){
			DrawGraph((elem.x - 1) * PIECEX, (elem.y - 1) * PIECEY, mImgPiece[elem.color], true);
		}
	}
	// ܂ł̃ubN`(ԕ邽FIELDɖ₢킹Ƃ1Z)
	for (int i = 0; i < BOARDY; ++i){
		for (int j = 0; j < BOARDX; ++j){
			if (mField[i + 1][j + 1]){
				DrawGraph(PIECEX * j, PIECEY * i, mImgPiece[mField[i + 1][j + 1]], true);
			}
		}
	}
	// Q[I[o[\
	if (mIsOver){
		DrawFormatString(60, 100, GetColor(0, 0, 0), "GAME OVER");
	}
#ifdef DEBUG
	DrawFormatString(0, 0, GetColor(0, 0, 0), "1:(%d,%d)", mNow[1].x, mNow[1].y);
#endif
}

//---------------------------------
// Q[s
void Game::Run()
{
	enum{
		INTERVAL_FRAME = 400,
	};

	++mFrame;
	if (!mIsOver){
		switch (mGameFlag){
		case GameFlag::Normal:
			this->KeyProcess();

			// t[1񎩓IɃubN
			if (mFrame % BLOCK_DOWN_FRAME == 0)
				this->BlockDown();
			break;

		case GameFlag::BlockDelete:
			if (GetNowCount() > mLastStateFrame + INTERVAL_FRAME){
				this->EraseAll();
			}
			break;
			
		case GameFlag::FieldDrop:
			if (GetNowCount() > mLastStateFrame + INTERVAL_FRAME){
				this->FieldDrop();
			}
			break;

		default:
			SHOULD_NOT_REACH_HERE;
		}
	}

	this->Draw();
}

//---------------------------------
// t[̏
bool LoopProc()
{
	return !ScreenFlip() && !ProcessMessage() && !ClearDrawScreen() && !CheckHitKey(KEY_INPUT_ESCAPE);
}

//---------------------------------------------------------
// Gg|Cg
//---------------------------------------------------------
int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
{
	ChangeWindowMode(true);
	SetGraphMode(PIECEX * BOARDX, PIECEY * BOARDY, 32);
	SetWindowText("Game");
	SetOutApplicationLogValidFlag(false);
	if (DxLib_Init()) return -1;

	try{
		while (LoopProc()){
			Key::get()->Update();
			Game::get()->Run();
		}
	}
	catch (const char* s){
		::MessageBox(nullptr, s, "Error", MB_OK | MB_ICONASTERISK);
	}
	catch (...){
		::MessageBox(nullptr, "unexpected exception raised", "Error", MB_OK | MB_ICONASTERISK);
	}

	DxLib_End();
	return 0;
}
